#include <xc.h>
#include <stdio.h>
#include "delay.h"
#include "main.h"
#include "lcd.h"
#include "uart.h"
// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = OFF       // Internal/External Switchover (Internal/External Switchover mode is disabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config VCAPEN = OFF     // Voltage Regulator Capacitor Enable bit (VCAP pin function disabled)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)

float voltage_battery = 0;
float voltage_temperature = 0;
float temperature = 0;
float temperature_avg = 0;
unsigned short temperature_adc = 0;
unsigned short temperature_mv = 0;
char backlight = 0;
char backlight_alwayson = 0;
unsigned short backlight_time = 0;

char button_counter = 0;

void main_init() {
    OSCCONbits.IRCF = 0x0F; /* raise internal oscillator clock to 16 Mhz        */
    while (OSCSTATbits.HFIOFR==0) {
        delay_1tcy();       /* wait until internal oscillator is ready          */
    }
    while (OSCSTATbits.HFIOFS==0) {
        delay_1tcy();       /* wait until oscillator stabilizes, within 0.5%    */
    }
    ANSELA = 0;    /* enable analogue input on PORT A, RA0 (v check)   */
    ANSELB = 0;             /* disable analogue operations on the other ports   */
    ANSELC = 0;
    ANSELD = 0;
    ANSELE = 0;
    TRISA  = 0b00000100;    /* set RA2 to input mode, for buttons    */
    TRISB  = 0;             /* set PORT B,C,D,E to output                       */
    TRISC  = 0;
    TRISD  = 0;

    FVRCONbits.ADFVR = FVR_4V;   /* Fixed Voltage Reference: 4.096v , 4mV per bit    */
    FVRCONbits.FVREN = 1;   /* Enable voltage reference                         */

/*  FVRRDY is always ‘1’ on PIC16F151X only. So commenting this.
 *   while (FVRCONbits.FVRRDY==0) {
 *       delay_1tcy();       // wait for the voltage reference to stabilize
 *   }
 */

    ADCON0bits.CHS = 0;     /* AN0 , battery voltage check */
    ADCON1bits.ADCS = 7;    /* 7 = internal oscillator */
    ADCON1bits.ADFM = 1;    /* [......xx] [xxxxxxxx] (right justified) */
    ADCON1bits.ADPREF = 3;  /* voltage reference set to internal reference */
    ADCON0bits.ADON = 1;    /* enable adc, leave it running all time */

    backlight = 1;
    backlight_time = 0;
    LCD_LIGHT = 1;
}



/* switches voltage reference between the three levels: 1.024, 2.048 , 4.096 
 * Note that the ADC only works with 2.048v or 4.096v settings on internal vref
 * Other options for ADC are external vref (1.8v or greater) or Vdd 
 */

void vref_set(char value) {
    if (FVRCONbits.ADFVR==value) return;
    if (value>3) return;
    FVRCONbits.FVREN = 0;
    FVRCONbits.ADFVR = value;
    FVRCONbits.FVREN = 1;
    /* on PIC16F1519 voltage reference is always "ready" due to the internal voltage
     * regulator the chip has, so in theory there's only switching between levels
     */
    while (FVRCONbits.FVRRDY==0) {
         /* wait for the voltage reference to stabilize   */
    }
    return;
}

char vref_get() {
    return FVRCONbits.ADFVR;
}

unsigned short adc_read(char id) {
    ADCON0bits.CHS = id;    /* set the channel on the A/D converter */
    //ADCON0bits.ADON = 1;  /* turn ON the ADC (always on, turned on in init */
    delay_25us();           /* wait for sampling capacitor to charge */
    delay_25us();
    ADCON0bits.ADGO = 1;    /* start the ADC process */
    while (ADCON0bits.ADGO == 1) {
        /* wait until conversion is done */
    }
    //ADCON0bits.ADON = 0;    /* turn off ADC (we leave it on non-stop */
    unsigned short nr = (ADRESHbits.ADRESH << 8);
    nr = nr + ADRESLbits.ADRESL;
    return nr;
}

/* Check the input voltage. Show battery low if voltage below a threshold
 *
 * AD595 runs directly from 9V battery, microcontroller and LCD display need 5v.
 * AD595 needs 2v above the maximum output we want from it, which is ~5.1v for
 * 500c, so we should show low battery when battery voltage goes down below ~7v.
 *
 * Microcontroller needs 4.75v to get a stable 4.096v reference. LCD needs 4.5v+
 * Typical linear regulators like 7805 need 1.5-2v above output voltage to work,
 * so showing a low battery voltage is also useful for this reason.
 *
 * Input voltage connected to PORTA:0 through a voltage divider defined by the
 * VOLTAGE_RATIO constant in temperature.h (about 4.75:1), therefore ADC will see
 * about 10v/4.75 = 2v for a new 9v battery.
 */

void voltage_get() {
    /* disable temperature input */
    TRISAbits.TRISA1 = 0;
    ANSELAbits.ANSA1 = 0;
    /* set pin AN0 to analogue input */
    TRISAbits.TRISA0 = 1;
    ANSELAbits.ANSA0 = 1;
    if (FVRCONbits.ADFVR == FVR_2V) vref_set(FVR_4V);
    unsigned short value = adc_read(ADC_VOLTAGE); /* read battery voltage       */

    TRISAbits.TRISA0 = 0; /* turn it off, we're done with it */
    ANSELAbits.ANSA0 = 0;

    /* measured voltage will be about 0.15-0.3v lower than battery/dc input due to
     * the 1N5818 diode in series with the input, with the role of reverse voltage
     * protection
     * If you want, feel free to add that value to the result, to match battery
     * or DC input voltage (even though it would not be accurate anymore)
     */

    voltage_battery = (value*4+2) * VOLTAGE_RATIO;
}

void temperature_get() {
    char multiplier = FVR_4V;
    unsigned short value;
    /* disable battery voltage input (should be turned off by the function, 
     * but better safe than sorry.
     */
    TRISAbits.TRISA0 = 0;
    TRISAbits.TRISA1 = 1;
    /* configure pin for analogue input */
    ANSELAbits.ANSA0 = 0;
    ANSELAbits.ANSA1 = 1;
    
    temperature_adc = adc_read(ADC_TEMPERATURE);
    multiplier = vref_get();
    if (multiplier==FVR_2V) {
        /* if vref is set to 2.048v and we're getting close to 2v, 
         * switch the reference to 4.096v 
         */
        if (value > 1010) vref_set(FVR_4V);
    } else {
        /* input voltage is less than 2v on the 0..4.096v scale, so we can 
         * switch the vref to 2.048v for better precision (in theory)
         */
        if (value < 450) vref_set(FVR_2V);
    }
    /* Convert ADC reading to voltage:
     * Default would be adc value * [2048|4096] mV / 1024 steps 
     * For better precision 0.5 adc count is added to the formula resulting in:
     * = (adc value + 0.5) * [2048|4096] mV / 1024 steps
     * Now we get rid of floating point by multiplying with 2:
     * = (2* adc value + 1) * [2048|4096] mV / (1024 steps * 2)
     * = (2*adc + 1) * [2048|4096] / 2048  || <- reduce by 2048
     * = (2*adc + 1) * [1|2]
     */
    value = 4*temperature_adc + 2;
    if (multiplier==FVR_2V) value = value >> 1;
    voltage_temperature = value * TEMPERATURE_RATIO;
    temperature = temperature_from_coefficients(voltage_temperature);
    if (TEMPERATURE_AVERAGE == 1 ) {
        /* averaging enabled, so pass it through the averaging function */
        temperature_avg = temperature_average(temperature);
    }

}

void voltage_show(char mode){
    char lcd_line[20];
    voltage_get();
    if (mode==0) { // startup
        sprintf(lcd_line,"     %2.3fv     ",voltage_battery/1000);
        lcd_writestring(lcd_line);
        if (voltage_battery < VOLTAGE_THRESHOLD) {
            lcd_goto(0,1);
            lcd_writestring("  BATTERY LOW  ");
        }
        delay_ms(200);
        delay_ms(200);
        delay_ms(200);
        delay_ms(200);
    }
    if (mode==1) {
        if (voltage_battery < VOLTAGE_THRESHOLD) {
            lcd_goto(0,0);
            lcd_writestring("Low battery:");
        }
        lcd_goto(12,0);
        sprintf(lcd_line,"%1.1fv",voltage_battery/1000);
        lcd_writestring(lcd_line);
    }

}

void temperature_show() {
    char lcd_line[20];
//    voltage_get();
//    lcd_goto(0,0);
//    sprintf(lcd_line,"          %2.2fv",voltage_battery/1000);
//    lcd_writestring(lcd_line);
    char spaces = 0;
    if (temperature <100) spaces = 1;
    if (temperature <10 ) spaces++;

    sprintf(lcd_line,"%.2fC",temperature);
    lcd_goto(0,1);
    if (spaces>0) {
        for (char i = 0;i<spaces;i++) lcd_writechar(' ');
    }
    lcd_writestring(lcd_line);

    char spaces = 0;
    if (temperature_avg <100) spaces = 1;
    if (temperature_avg <10 ) spaces++;
    lcd_writestring("  ");
    if (spaces>0) {
        for (char i = 0;i<spaces;i++) lcd_writechar(' ');
    }
    sprintf(lcd_line,"%.2fC",temperature_avg);
    lcd_writestring(lcd_line);
    delay_ms(TEMPERATURE_DISPLAYTIME);
}

void uart_log() {
    char lcd_line[80];
    sprintf(lcd_line,"temp=\t%3.3f\tadc=\t%d\tvbat=\t%2.2f\t\r\n",temperature, temperature_adc, voltage_battery);
    uart_writestring(lcd_line);
}


void button_pressed(char value) {
    if (value==1) {  // small button press
        if (backlight_alwayson == 0) {
            LCD_LIGHT = 1;
            backlight_time = 1;
            return;
       }
    }
    if (value==2) { // long button press
        if (backlight_alwayson == 0 ) {
            backlight_alwayson = 1;
            backlight_time = 0;
            LCD_LIGHT = 1;
        } else {
            backlight_alwayson = 0;
            backlight_time = 0;
            LCD_LIGHT = 0;
        }
    }
}

void main() {
    main_init();
    lcd_init();
    uart_init();
    lcd_home();
    // turn backlight on by faking a button press, so user will see battery voltage
    button_pressed(1);
    voltage_show(0); // it's showing on whole screen for 1s or so
    lcd_clear();
    voltage_show(1); // now we clean the display and show it properly on side
    
    unsigned char battery_counter = VOLTAGE_REFRESH;
    while (1) {
        temperature_get();
        temperature_show();
        PORTCbits.RC3 = 1;
        uart_log();
        PORTCbits.RC3 = 0;
        battery_counter--;
        if (battery_counter == 0) {
            battery_counter = VOLTAGE_REFRESH;
            voltage_get();
            voltage_show(1);
        }
        if (backlight_alwayson==0) {
            if (backlight_time !=0) {
                backlight_time++;
                if (backlight_time > BACKLIGHT_TIMEOUT) {
                    backlight_time = 0;
                    LCD_LIGHT = 0;
                }
            }
        }
        if (BUTTON_LIGHT == 0) {
            if (button_counter == 255) { // it was a long button press, reset it
                button_counter = 0;
            } else {
                if (button_counter > 0) {
                    button_counter = 0;
                    button_pressed(1);
                }
            }
        }
        if (BUTTON_LIGHT == 1) {
            if (button_counter !=255) {
                button_counter++;
                if (button_counter >= BUTTON_LONGPRESS) {
                    button_counter = 255;
                    button_pressed(2);
                }
            }
        }

    } /* while */
}

