#include "main.h";



/* From http://srdata.nist.gov/its90/download/type_k.tab
* ************************************
* This section contains coefficients of approximate inverse
* functions for type K thermocouples for the subranges of
* temperature and voltage listed below.
*
* The coefficients are in units of °C and mV and are listed in
* the order of constant term up to the highest order.
* The equation is of the form t_90 = d_0 + d_1*E + d_2*E^2 + ...  + d_n*E^n,
* where E is in mV and t_90 is in °C.
*
*    Temperature        Voltage            Error
*      range              range            range
*      (°C)               (mV)             (° C)
*    -200. to 0.      -5.891 to 0.000    -0.02 to 0.04
*     0. to 500.      0.000 to 20.644    -0.05 to 0.04
*     500. to 1372.   20.644 to 54.886   -0.05 to 0.06
********************************************************
Inverse coefficients for type K:

Temperature  -200.             0.           500.
  Range:        0.           500.          1372.

  Voltage   -5.891          0.000         20.644
  Range:     0.000         20.644         54.886

         0.0000000E+00  0.000000E+00 -1.318058E+02
         2.5173462E+01  2.508355E+01  4.830222E+01
        -1.1662878E+00  7.860106E-02 -1.646031E+00
        -1.0833638E+00 -2.503131E-01  5.464731E-02
        -8.9773540E-01  8.315270E-02 -9.650715E-04
        -3.7342377E-01 -1.228034E-02  8.802193E-06
        -8.6632643E-02  9.804036E-04 -3.110810E-08
        -1.0450598E-02 -4.413030E-05  0.000000E+00
        -5.1920577E-04  1.057734E-06  0.000000E+00
         0.0000000E+00 -1.052755E-08  0.000000E+00

  Error      -0.02          -0.05          -0.05
  Range:      0.04           0.04           0.06
 */

const float coefficients[3][10] = { {0, 25.173462, -1.1662878, -1.0833638, -0.8977354, -0.37342377, -0.086632643, -0.010450598, -0.00051920577, 0},
                                    {0, 25.08355, 0.07860106, -0.2503131, 0.08315270, -0.01228034, 0.0009804036, -0.0000441303, 0.000001057734, -0.00000001052755},
                                    {-131.8058, 48.30222, -1.646031, 0.05464731, -0.0009650715, 0.000008802193, -0.0000000311081, 0, 0, 0},
                                  };


float tavg_values[TEMPERATURE_BUFFERSIZE];
float tavg_sum = 0;
float tavg_prev = 0;
char tavg_init = 0;
char tavg_oldest = 0; /* the oldest temperature offset that's going to be removed from array when new one comes */
float t_current = 0;


/* Accepts a new floating point temperature and returns a new average based on 
 * previously memorized values. Oldest temperature is removed and replaced with
 * this new received temperature.
  */
float temperature_average(float temperature_new) {
    
    /* t_current is (a very close approximation of the) sum of all
     * previously recorded temperatures.
     * Remove oldest, store new temperature in its place, mark the next
     * temperature stored in array as oldest (or reset to first in array if
     * the end of array is reached
     */

    if (tavg_init==0) {
        for (char i=0;i<TEMPERATURE_BUFFERSIZE;i++) {
            tavg_values[i] = temperature_new;
        }
        tavg_sum = temperature_new * TEMPERATURE_BUFFERSIZE;
        tavg_oldest = 0;
        tavg_prev = 0;
        tavg_init = 1;
        return temperature_new;
    }

    tavg_sum = tavg_sum - tavg_values[tavg_oldest] + temperature_new;
    tavg_values[tavg_oldest] = temperature_new;


    t_current = t_current - tavg_values[tavg_oldest];
    tavg_values[tavg_oldest] = temperature_new;
    tavg_oldest++;
    if (tavg_oldest >= TEMPERATURE_BUFFERSIZE) {
        tavg_oldest = 0;
    }
    return tavg_sum / TEMPERATURE_BUFFERSIZE;

}

float temperature_from_coefficients(float value) {
char idx = 0;
float v = (value / 247.3) - 0.011; /* thermocouple voltage */
if (v < 0) return 0;
if (v<=20.644) idx=1; // pick 0..500c coefficients
if (v>20.644) idx=2;  // pick 500c..1300c coefficients

/* formula is d0 + d1*v + d2*v^2 + d3*v^3 + .. + d9*v^9
 *
 * we can rewrite it like this to compute this easier:
 *
 *  d0 + d1*v + d2*v^2  =>  d0 + v*(d1 + d2*v)
 *
 * d0 + v(d1+v(d2+v*(d3+v*(d4+v*(d5+v*(d6+v*(d7+v*(d8+v*d9)))))))))...)
 *
 * so we go backwards 
  */
float temp = coefficients[idx][9];
for (char i =9; i>0;i--) {
    temp = temp*v + coefficients[idx][i-1];
}
return temp;
}